# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import concurrent.futures
import math
from typing import Any, Dict, Union

import numpy as np

from braket.aws import AwsDevice
from braket.circuits import Circuit, Gate, Instruction, Observable, ResultType
from braket.circuits.quantum_operator_helpers import get_pauli_eigenvalues
from braket.circuits.serialization import IRType
from braket.devices import Device
from braket.ir.openqasm import Program as OpenQasmProgram
from braket.tasks import GateModelQuantumTaskResult


def get_tol(shots: int) -> Dict[str, float]:
    return {"atol": 0.2, "rtol": 0.3} if shots else {"atol": 0.01, "rtol": 0}


def qubit_ordering_testing(device: Device, run_kwargs: Dict[str, Any]):
    # |110> should get back value of "110"
    state_110 = Circuit().x(0).x(1).i(2)
    result = device.run(state_110, **run_kwargs).result()
    assert result.measurement_counts.most_common(1)[0][0] == "110"
    state_110_qasm = state_110.to_ir(ir_type=IRType.OPENQASM)
    result = device.run(state_110_qasm, **run_kwargs).result()
    assert result.measurement_counts.most_common(1)[0][0] == "110"

    # |001> should get back value of "001"
    state_001 = Circuit().i(0).i(1).x(2)
    result = device.run(state_001, **run_kwargs).result()
    assert result.measurement_counts.most_common(1)[0][0] == "001"
    state_001_qasm = state_001.to_ir(ir_type=IRType.OPENQASM)
    result = device.run(state_001_qasm, **run_kwargs).result()
    assert result.measurement_counts.most_common(1)[0][0] == "001"


def no_result_types_testing(
    program: Union[Circuit, OpenQasmProgram],
    device: Device,
    run_kwargs: Dict[str, Any],
    expected: Dict[str, float],
):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    result = device.run(program, **run_kwargs).result()
    probabilities = result.measurement_probabilities
    for bitstring in probabilities:
        assert np.allclose(probabilities[bitstring], expected[bitstring], **tol)
    assert len(result.measurements) == shots


def no_result_types_bell_pair_testing(device: Device, run_kwargs: Dict[str, Any]):
    bell = Circuit().h(0).cnot(0, 1)
    bell_qasm = bell.to_ir(ir_type=IRType.OPENQASM)
    for task in (bell, bell_qasm):
        no_result_types_testing(task, device, run_kwargs, {"00": 0.5, "11": 0.5})


def result_types_observable_not_in_instructions(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    bell = (
        Circuit()
        .h(0)
        .cnot(0, 1)
        .expectation(observable=Observable.X(), target=[2])
        .variance(observable=Observable.Y(), target=[3])
    )
    bell_qasm = bell.to_ir(ir_type=IRType.OPENQASM)
    results = device.run_batch([bell, bell_qasm], **run_kwargs).results()
    for result in results:
        assert np.allclose(result.values[0], 0, **tol)
        assert np.allclose(result.values[1], 1, **tol)


def result_types_zero_shots_bell_pair_testing(
    device: Device,
    include_state_vector: bool,
    run_kwargs: Dict[str, Any],
    include_amplitude: bool = True,
):
    circuit = (
        Circuit()
        .h(0)
        .cnot(0, 1)
        .expectation(observable=Observable.H() @ Observable.X(), target=[0, 1])
    )
    if include_amplitude:
        circuit.amplitude(["01", "10", "00", "11"])
    if include_state_vector:
        circuit.state_vector()
    tasks = [circuit, circuit.to_ir(ir_type=IRType.OPENQASM)]
    results = device.run_batch(tasks, **run_kwargs).results()
    for result in results:
        assert len(result.result_types) == 3 if include_state_vector else 2
        assert np.allclose(
            result.get_value_by_result_type(
                ResultType.Expectation(observable=Observable.H() @ Observable.X(), target=[0, 1])
            ),
            1 / np.sqrt(2),
        )
        if include_state_vector:
            assert np.allclose(
                result.get_value_by_result_type(ResultType.StateVector()),
                np.array([1, 0, 0, 1]) / np.sqrt(2),
            )
        if include_amplitude:
            amplitude = result.get_value_by_result_type(
                ResultType.Amplitude(["01", "10", "00", "11"])
            )
            assert np.isclose(amplitude["01"], 0)
            assert np.isclose(amplitude["10"], 0)
            assert np.isclose(amplitude["00"], 1 / np.sqrt(2))
            assert np.isclose(amplitude["11"], 1 / np.sqrt(2))


def result_types_bell_pair_full_probability_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    circuit = Circuit().h(0).cnot(0, 1).probability()
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()
        assert len(result.result_types) == 1
        assert np.allclose(
            result.get_value_by_result_type(ResultType.Probability()),
            np.array([0.5, 0, 0, 0.5]),
            **tol,
        )


def result_types_bell_pair_marginal_probability_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    circuit = Circuit().h(0).cnot(0, 1).probability(0)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()
        assert len(result.result_types) == 1
        assert np.allclose(
            result.get_value_by_result_type(ResultType.Probability(target=0)),
            np.array([0.5, 0.5]),
            **tol,
        )


def result_types_nonzero_shots_bell_pair_testing(device: Device, run_kwargs: Dict[str, Any]):
    circuit = (
        Circuit()
        .h(0)
        .cnot(0, 1)
        .expectation(observable=Observable.H() @ Observable.X(), target=[0, 1])
        .sample(observable=Observable.H() @ Observable.X(), target=[0, 1])
    )
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()
        assert len(result.result_types) == 2
        assert (
            0.6
            < result.get_value_by_result_type(
                ResultType.Expectation(observable=Observable.H() @ Observable.X(), target=[0, 1])
            )
            < 0.8
        )
        assert (
            len(
                result.get_value_by_result_type(
                    ResultType.Sample(observable=Observable.H() @ Observable.X(), target=[0, 1])
                )
            )
            == run_kwargs["shots"]
        )


def result_types_hermitian_testing(
    device: Device, run_kwargs: Dict[str, Any], test_program: bool = True
):
    shots = run_kwargs["shots"]
    theta = 0.543
    array = np.array([[1, 2j], [-2j, 0]])

    circuit = (
        Circuit()
        .rx(0, theta)
        .variance(Observable.Hermitian(array), 0)
        .expectation(Observable.Hermitian(array), 0)
    )
    if shots:
        circuit.add_result_type(ResultType.Sample(Observable.Hermitian(array), 0))
    tasks = (circuit,) if not test_program else (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = 2 * np.sin(theta) + 0.5 * np.cos(theta) + 0.5
        expected_var = 0.25 * (np.sin(theta) - 4 * np.cos(theta)) ** 2
        expected_eigs = np.linalg.eigvalsh(array)
        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_all_selected_testing(
    device: Device, run_kwargs: Dict[str, Any], test_program: bool = True
):
    shots = run_kwargs["shots"]
    theta = 0.543
    array = np.array([[1, 2j], [-2j, 0]])

    circuit = (
        Circuit()
        .rx(0, theta)
        .rx(1, theta)
        .variance(Observable.Hermitian(array))
        .expectation(Observable.Hermitian(array), 0)
    )
    if shots:
        circuit.add_result_type(ResultType.Sample(Observable.Hermitian(array), 1))

    tasks = (circuit,) if not test_program else (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))

    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = 2 * np.sin(theta) + 0.5 * np.cos(theta) + 0.5
        var = 0.25 * (np.sin(theta) - 4 * np.cos(theta)) ** 2
        expected_var = [var, var]
        expected_eigs = np.linalg.eigvalsh(array)
        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots) -> Circuit:
    circuit = (
        Circuit()
        .rx(0, theta)
        .rx(1, phi)
        .rx(2, varphi)
        .cnot(0, 1)
        .cnot(1, 2)
        .variance(obs, obs_targets)
        .expectation(obs, obs_targets)
    )
    if shots:
        circuit.sample(obs, obs_targets)
    return circuit


def assert_variance_expectation_sample_result(
    result: GateModelQuantumTaskResult,
    shots: int,
    expected_var: float,
    expected_mean: float,
    expected_eigs: np.ndarray,
):
    tol = get_tol(shots)
    variance = result.values[0]
    expectation = result.values[1]
    if shots:
        samples = result.values[2]
        assert np.allclose(sorted(list(set(samples))), sorted(expected_eigs), **tol)
        assert np.allclose(np.mean(samples), expected_mean, **tol)
        assert np.allclose(np.var(samples), expected_var, **tol)
    assert np.allclose(expectation, expected_mean, **tol)
    assert np.allclose(variance, expected_var, **tol)


def result_types_tensor_x_y_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    obs = Observable.X() @ Observable.Y()
    obs_targets = [0, 2]
    circuit = get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = np.sin(theta) * np.sin(phi) * np.sin(varphi)
        expected_var = (
            8 * np.sin(theta) ** 2 * np.cos(2 * varphi) * np.sin(phi) ** 2
            - np.cos(2 * (theta - phi))
            - np.cos(2 * (theta + phi))
            + 2 * np.cos(2 * theta)
            + 2 * np.cos(2 * phi)
            + 14
        ) / 16
        expected_eigs = get_pauli_eigenvalues(1)

        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_tensor_z_z_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    obs = Observable.Z() @ Observable.Z()
    obs_targets = [0, 2]
    circuit = get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = 0.849694136476246
        expected_var = 0.27801987443788634
        expected_eigs = get_pauli_eigenvalues(1)

        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_tensor_hermitian_hermitian_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    matrix1 = np.array([[1, 2], [2, 4]])
    matrix2 = np.array(
        [
            [-6, 2 + 1j, -3, -5 + 2j],
            [2 - 1j, 0, 2 - 1j, -5 + 4j],
            [-3, 2 + 1j, 0, -4 + 3j],
            [-5 - 2j, -5 - 4j, -4 - 3j, -6],
        ]
    )
    obs = Observable.Hermitian(matrix1) @ Observable.Hermitian(matrix2)
    obs_targets = [0, 1, 2]
    circuit = get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = -4.30215023196904
        expected_var = 370.71292282796804
        expected_eigs = np.array([-70.90875406, -31.04969387, 0, 3.26468993, 38.693758])

        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_tensor_z_h_y_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    obs = Observable.Z() @ Observable.H() @ Observable.Y()
    obs_targets = [0, 1, 2]
    circuit = get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = -(np.cos(varphi) * np.sin(phi) + np.sin(varphi) * np.cos(theta)) / np.sqrt(
            2
        )
        expected_var = (
            3
            + np.cos(2 * phi) * np.cos(varphi) ** 2
            - np.cos(2 * theta) * np.sin(varphi) ** 2
            - 2 * np.cos(theta) * np.sin(phi) * np.sin(2 * varphi)
        ) / 4
        expected_eigs = get_pauli_eigenvalues(1)
        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_tensor_z_hermitian_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    array = np.array(
        [
            [-6, 2 + 1j, -3, -5 + 2j],
            [2 - 1j, 0, 2 - 1j, -5 + 4j],
            [-3, 2 + 1j, 0, -4 + 3j],
            [-5 - 2j, -5 - 4j, -4 - 3j, -6],
        ]
    )
    obs = Observable.Z() @ Observable.Hermitian(array)
    obs_targets = [0, 1, 2]
    circuit = get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = 0.5 * (
            -6 * np.cos(theta) * (np.cos(varphi) + 1)
            - 2 * np.sin(varphi) * (np.cos(theta) + np.sin(phi) - 2 * np.cos(phi))
            + 3 * np.cos(varphi) * np.sin(phi)
            + np.sin(phi)
        )
        expected_var = (
            1057
            - np.cos(2 * phi)
            + 12 * (27 + np.cos(2 * phi)) * np.cos(varphi)
            - 2 * np.cos(2 * varphi) * np.sin(phi) * (16 * np.cos(phi) + 21 * np.sin(phi))
            + 16 * np.sin(2 * phi)
            - 8 * (-17 + np.cos(2 * phi) + 2 * np.sin(2 * phi)) * np.sin(varphi)
            - 8 * np.cos(2 * theta) * (3 + 3 * np.cos(varphi) + np.sin(varphi)) ** 2
            - 24 * np.cos(phi) * (np.cos(phi) + 2 * np.sin(phi)) * np.sin(2 * varphi)
            - 8
            * np.cos(theta)
            * (
                4
                * np.cos(phi)
                * (
                    4
                    + 8 * np.cos(varphi)
                    + np.cos(2 * varphi)
                    - (1 + 6 * np.cos(varphi)) * np.sin(varphi)
                )
                + np.sin(phi)
                * (
                    15
                    + 8 * np.cos(varphi)
                    - 11 * np.cos(2 * varphi)
                    + 42 * np.sin(varphi)
                    + 3 * np.sin(2 * varphi)
                )
            )
        ) / 16

        z_array = np.diag([1, -1])
        expected_eigs = np.linalg.eigvalsh(np.kron(z_array, array))
        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_tensor_y_hermitian_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    array = np.array(
        [
            [-6, 2 + 1j, -3, -5 + 2j],
            [2 - 1j, 0, 2 - 1j, -5 + 4j],
            [-3, 2 + 1j, 0, -4 + 3j],
            [-5 - 2j, -5 - 4j, -4 - 3j, -6],
        ]
    )
    obs = Observable.Y() @ Observable.Hermitian(array)
    obs_targets = [0, 1, 2]
    circuit = get_result_types_three_qubit_circuit(theta, phi, varphi, obs, obs_targets, shots)
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean = 1.4499810303182408
        expected_var = 74.03174647518193
        y_array = np.array([[0, -1j], [1j, 0]])
        expected_eigs = np.linalg.eigvalsh(np.kron(y_array, array))
        assert_variance_expectation_sample_result(
            result, shots, expected_var, expected_mean, expected_eigs
        )


def result_types_noncommuting_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = 0
    theta = 0.432
    phi = 0.123
    varphi = -0.543
    array = np.array(
        [
            [-6, 2 + 1j, -3, -5 + 2j],
            [2 - 1j, 0, 2 - 1j, -5 + 4j],
            [-3, 2 + 1j, 0, -4 + 3j],
            [-5 - 2j, -5 - 4j, -4 - 3j, -6],
        ]
    )
    obs1 = Observable.X() @ Observable.Y()
    obs1_targets = [0, 2]
    obs2 = Observable.Z() @ Observable.Z()
    obs2_targets = [0, 2]
    obs3 = Observable.Y() @ Observable.Hermitian(array)
    obs3_targets = [0, 1, 2]
    circuit = (
        get_result_types_three_qubit_circuit(theta, phi, varphi, obs1, obs1_targets, shots)
        .expectation(obs2, obs2_targets)
        .expectation(obs3, obs3_targets)
    )
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()

        expected_mean1 = np.sin(theta) * np.sin(phi) * np.sin(varphi)
        expected_var1 = (
            8 * np.sin(theta) ** 2 * np.cos(2 * varphi) * np.sin(phi) ** 2
            - np.cos(2 * (theta - phi))
            - np.cos(2 * (theta + phi))
            + 2 * np.cos(2 * theta)
            + 2 * np.cos(2 * phi)
            + 14
        ) / 16

        expected_mean2 = 0.849694136476246
        expected_mean3 = 1.4499810303182408
        assert np.allclose(result.values[0], expected_var1)
        assert np.allclose(result.values[1], expected_mean1)
        assert np.allclose(result.values[2], expected_mean2)
        assert np.allclose(result.values[3], expected_mean3)


def result_types_noncommuting_flipped_targets_testing(device: Device, run_kwargs: Dict[str, Any]):
    circuit = (
        Circuit()
        .h(0)
        .cnot(0, 1)
        .expectation(observable=Observable.H() @ Observable.X(), target=[0, 1])
        .expectation(observable=Observable.H() @ Observable.X(), target=[1, 0])
    )
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, shots=0, **run_kwargs).result()
        assert np.allclose(result.values[0], np.sqrt(2) / 2)
        assert np.allclose(result.values[1], np.sqrt(2) / 2)


def result_types_noncommuting_all(device: Device, run_kwargs: Dict[str, Any]):
    array = np.array([[1, 2j], [-2j, 0]])
    circuit = (
        Circuit()
        .h(0)
        .cnot(0, 1)
        .expectation(observable=Observable.Hermitian(array))
        .expectation(observable=Observable.X())
    )
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, shots=0, **run_kwargs).result()
        assert np.allclose(result.values[0], [0.5, 0.5])
        assert np.allclose(result.values[1], [0, 0])


def multithreaded_bell_pair_testing(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    bell = Circuit().h(0).cnot(0, 1)

    def run_circuit(circuit):
        task = device.run(circuit, **run_kwargs)
        return task.result()

    futures = []
    num_threads = 2

    tasks = (bell, bell.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for _ in range(num_threads):
                future = executor.submit(run_circuit, task)
                futures.append(future)
        for future in futures:
            result = future.result()
            assert np.allclose(result.measurement_probabilities["00"], 0.5, **tol)
            assert np.allclose(result.measurement_probabilities["11"], 0.5, **tol)
            assert len(result.measurements) == shots


def noisy_circuit_1qubit_noise_full_probability(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    circuit = Circuit().x(0).x(1).bit_flip(0, 0.1).probability()
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()
        assert len(result.result_types) == 1
        assert np.allclose(
            result.get_value_by_result_type(ResultType.Probability()),
            np.array([0.0, 0.1, 0, 0.9]),
            **tol,
        )


def noisy_circuit_2qubit_noise_full_probability(device: Device, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    K0 = np.eye(4) * np.sqrt(0.9)
    K1 = np.kron(np.array([[0.0, 1.0], [1.0, 0.0]]), np.array([[0.0, 1.0], [1.0, 0.0]])) * np.sqrt(
        0.1
    )
    circuit = Circuit().x(0).x(1).kraus((0, 1), [K0, K1]).probability()
    tasks = (circuit, circuit.to_ir(ir_type=IRType.OPENQASM))
    for task in tasks:
        result = device.run(task, **run_kwargs).result()
        assert len(result.result_types) == 1
        assert np.allclose(
            result.get_value_by_result_type(ResultType.Probability()),
            np.array([0.1, 0.0, 0, 0.9]),
            **tol,
        )


def batch_bell_pair_testing(device: AwsDevice, run_kwargs: Dict[str, Any]):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    circuits = [Circuit().h(0).cnot(0, 1) for _ in range(10)]
    tasks_list = (circuits, [circuit.to_ir(ir_type=IRType.OPENQASM) for circuit in circuits])
    for tasks in tasks_list:
        batch = device.run_batch(tasks, max_parallel=5, **run_kwargs)
        results = batch.results()
        for result in results:
            assert np.allclose(result.measurement_probabilities["00"], 0.5, **tol)
            assert np.allclose(result.measurement_probabilities["11"], 0.5, **tol)
            assert len(result.measurements) == shots
        assert [task.result() for task in batch.tasks] == results


def bell_pair_openqasm_testing(device: AwsDevice, run_kwargs: Dict[str, Any]):
    openqasm_string = (
        "OPENQASM 3;"
        "qubit[2] q;"
        "bit[2] c;"
        "h q[0];"
        "cnot q[0], q[1];"
        "c[0] = measure q[0];"
        "c[1] = measure q[1];"
    )
    hardcoded_openqasm = OpenQasmProgram(source=openqasm_string)
    circuit = Circuit().h(0).cnot(0, 1)
    generated_openqasm = circuit.to_ir(ir_type=IRType.OPENQASM)

    for program in hardcoded_openqasm, generated_openqasm:
        no_result_types_testing(program, device, run_kwargs, {"00": 0.5, "11": 0.5})


def openqasm_noisy_circuit_1qubit_noise_full_probability(
    device: Device, run_kwargs: Dict[str, Any]
):
    shots = run_kwargs["shots"]
    tol = get_tol(shots)
    openqasm_string = (
        "OPENQASM 3;"
        "qubit[2] q;"
        "x q[0];"
        "x q[1];"
        "#pragma braket noise bit_flip(0.1) q[0]"
        "#pragma braket result probability q[0], q[1]"
    )
    hardcoded_openqasm = OpenQasmProgram(source=openqasm_string)
    circuit = Circuit().x(0).x(1).bit_flip(0, 0.1).probability([0, 1])
    generated_openqasm = circuit.to_ir(ir_type=IRType.OPENQASM)

    for program in hardcoded_openqasm, generated_openqasm:
        result = device.run(program, **run_kwargs).result()
        assert len(result.result_types) == 1
        assert np.allclose(
            result.get_value_by_result_type(ResultType.Probability(target=[0, 1])),
            np.array([0.0, 0.1, 0, 0.9]),
            **tol,
        )


def openqasm_result_types_bell_pair_testing(device: Device, run_kwargs: Dict[str, Any]):
    openqasm_string = (
        "OPENQASM 3;"
        "qubit[2] q;"
        "h q[0];"
        "cnot q[0], q[1];"
        "#pragma braket result expectation h(q[0]) @ x(q[1])"
        "#pragma braket result sample h(q[0]) @ x(q[1])"
    )
    hardcoded_openqasm = OpenQasmProgram(source=openqasm_string)
    circuit = (
        Circuit()
        .h(0)
        .cnot(0, 1)
        .expectation(Observable.H() @ Observable.X(), (0, 1))
        .sample(Observable.H() @ Observable.X(), (0, 1))
    )
    generated_openqasm = circuit.to_ir(ir_type=IRType.OPENQASM)

    for program in hardcoded_openqasm, generated_openqasm:
        result = device.run(program, **run_kwargs).result()
        assert len(result.result_types) == 2
        assert (
            0.6
            < result.get_value_by_result_type(
                ResultType.Expectation(observable=Observable.H() @ Observable.X(), target=[0, 1])
            )
            < 0.8
        )
        assert (
            len(
                result.get_value_by_result_type(
                    ResultType.Sample(observable=Observable.H() @ Observable.X(), target=[0, 1])
                )
            )
            == run_kwargs["shots"]
        )


def many_layers(n_qubits: int, n_layers: int) -> Circuit:
    """
    Function to return circuit with many layers.

    :param int n_qubits: number of qubits
    :param int n_layers: number of layers
    :return: Constructed easy circuit
    :rtype: Circuit
    """
    qubits = range(n_qubits)
    circuit = Circuit()  # instantiate circuit object
    for q in range(n_qubits):
        circuit.h(q)
    for layer in range(n_layers):
        if (layer + 1) % 100 != 0:
            for qubit in range(len(qubits)):
                angle = np.random.uniform(0, 2 * math.pi)
                gate = np.random.choice(
                    [Gate.Rx(angle), Gate.Ry(angle), Gate.Rz(angle), Gate.H()], 1, replace=True
                )[0]
                circuit.add_instruction(Instruction(gate, qubit))
        else:
            for q in range(0, n_qubits, 2):
                circuit.cnot(q, q + 1)
            for q in range(1, n_qubits - 1, 2):
                circuit.cnot(q, q + 1)
    return circuit



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import itertools
from collections.abc import Iterable
from typing import Any, ClassVar, Union

import numpy as np

import braket.ir.jaqcd as ir
from braket.circuits import circuit
from braket.circuits.free_parameter import FreeParameter
from braket.circuits.free_parameter_expression import FreeParameterExpression
from braket.circuits.gates import format_complex
from braket.circuits.instruction import Instruction
from braket.circuits.noise import (
    DampingNoise,
    GeneralizedAmplitudeDampingNoise,
    MultiQubitPauliNoise,
    Noise,
    PauliNoise,
    SingleProbabilisticNoise,
    SingleProbabilisticNoise_34,
    SingleProbabilisticNoise_1516,
)
from braket.circuits.quantum_operator_helpers import (
    is_cptp,
    verify_quantum_operator_matrix_dimensions,
)
from braket.circuits.serialization import OpenQASMSerializationProperties
from braket.registers.qubit import QubitInput
from braket.registers.qubit_set import QubitSet, QubitSetInput

"""
To add a new Noise implementation:
    1. Implement the class and extend `Noise`
    2. Add a method with the `@circuit.subroutine(register=True)` decorator. Method name
       will be added into the `Circuit` class. This method is the default way clients add
       this noise to a circuit.
    3. Register the class with the `Noise` class via `Noise.register_noise()`.
"""


class BitFlip(SingleProbabilisticNoise):
    r"""Bit flip noise channel which transforms a density matrix :math:`\\rho` according to:

    .. math:: \\rho \\Rightarrow (1-p) \\rho + p X \\rho X^{\\dagger}

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        X = \\left(
                \\begin{matrix}
                    0 & 1 \\\\
                    1 & 0
                \\end{matrix}
            \\right)

        p = \\text{probability}

    This noise channel is shown as `BF` in circuit diagrams.
    """

    def __init__(self, probability: Union[FreeParameterExpression, float]):
        super().__init__(
            probability=probability,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("BF", [probability])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.BitFlip.construct(target=target[0], probability=self.probability)

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return f"#pragma braket noise bit_flip({self.probability}) {target_qubit}"

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.sqrt(1 - self.probability) * np.eye(2, dtype=complex)
        K1 = np.sqrt(self.probability) * np.array([[0.0, 1.0], [1.0, 0.0]], dtype=complex)
        return [K0, K1]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def bit_flip(target: QubitSetInput, probability: float) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s)
            probability (float): Probability of bit flipping.

        Returns:
            Iterable[Instruction]: `Iterable` of BitFlip instructions.

        Examples:
            >>> circ = Circuit().bit_flip(0, probability=0.1)
        """
        return [
            Instruction(Noise.BitFlip(probability=probability), target=qubit)
            for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs: Union[FreeParameter, str]) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Args:
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return BitFlip(probability=_substitute_value(self._probability, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return BitFlip(probability=_parameter_from_dict(noise["probability"]))


Noise.register_noise(BitFlip)


class PhaseFlip(SingleProbabilisticNoise):
    r"""Phase flip noise channel which transforms a density matrix :math:`\\rho` according to:

    .. math:: \\rho \\Rightarrow (1-p) \\rho + p X \\rho X^{\\dagger}

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        Z = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & -1
                \\end{matrix}
            \\right)

        p = \\text{probability}

    This noise channel is shown as `PF` in circuit diagrams.
    """

    def __init__(self, probability: Union[FreeParameterExpression, float]):
        super().__init__(
            probability=probability,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("PF", [probability])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.PhaseFlip.construct(target=target[0], probability=self.probability)

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return f"#pragma braket noise phase_flip({self.probability}) {target_qubit}"

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.sqrt(1 - self.probability) * np.eye(2, dtype=complex)
        K1 = np.sqrt(self.probability) * np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex)
        return [K0, K1]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def phase_flip(target: QubitSetInput, probability: float) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s)
            probability (float): Probability of phase flipping.

        Returns:
            Iterable[Instruction]: `Iterable` of PhaseFlip instructions.

        Examples:
            >>> circ = Circuit().phase_flip(0, probability=0.1)
        """
        return [
            Instruction(Noise.PhaseFlip(probability=probability), target=qubit)
            for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs: Union[FreeParameter, str]) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Args:
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return PhaseFlip(probability=_substitute_value(self._probability, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return PhaseFlip(probability=_parameter_from_dict(noise["probability"]))


Noise.register_noise(PhaseFlip)


class PauliChannel(PauliNoise):
    r"""Pauli noise channel which transforms a density matrix :math:`\\rho` according to:

    .. math::
        \\rho \\Rightarrow (1-probX-probY-probZ) \\rho
            + probX X \\rho X^{\\dagger}
            + probY Y \\rho Y^{\\dagger}
            + probZ Z \\rho Z^{\\dagger}

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        X = \\left(
                \\begin{matrix}
                    0 & 1 \\\\
                    1 & 0
                \\end{matrix}
            \\right)

        Y = \\left(
                \\begin{matrix}
                    0 & -i \\\\
                    i &  0
                \\end{matrix}
            \\right)

        Z = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & -1
                \\end{matrix}
            \\right)

    This noise channel is shown as `PC` in circuit diagrams.
    """

    def __init__(
        self,
        probX: Union[FreeParameterExpression, float],
        probY: Union[FreeParameterExpression, float],
        probZ: Union[FreeParameterExpression, float],
    ):
        """Creates PauliChannel noise.

        Args:
            probX (Union[FreeParameterExpression, float]): X rotation probability.
            probY (Union[FreeParameterExpression, float]): Y rotation probability.
            probZ (Union[FreeParameterExpression, float]): Z rotation probability.
        """
        super().__init__(
            probX=probX,
            probY=probY,
            probZ=probZ,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("PC", [probX, probY, probZ])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.PauliChannel.construct(
            target=target[0], probX=self.probX, probY=self.probY, probZ=self.probZ
        )

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return (
            f"#pragma braket noise pauli_channel"
            f"({self.probX}, {self.probY}, {self.probZ}) {target_qubit}"
        )

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.sqrt(1 - self.probX - self.probY - self.probZ) * np.eye(2, dtype=complex)
        K1 = np.sqrt(self.probX) * np.array([[0.0, 1.0], [1.0, 0.0]], dtype=complex)
        K2 = np.sqrt(self.probY) * 1j * np.array([[0.0, -1.0], [1.0, 0.0]], dtype=complex)
        K3 = np.sqrt(self.probZ) * np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex)
        return [K0, K1, K2, K3]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def pauli_channel(
        target: QubitSetInput, probX: float, probY: float, probZ: float
    ) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s)
                probability list[float]: Probabilities for the Pauli X, Y and Z noise
                happening in the Kraus channel.
            probX (float): X rotation probability.
            probY (float): Y rotation probability.
            probZ (float): Z rotation probability.

        Returns:
            Iterable[Instruction]: `Iterable` of PauliChannel instructions.

        Examples:
            >>> circ = Circuit().pauli_channel(0, probX=0.1, probY=0.2, probZ=0.3)
        """
        return [
            Instruction(Noise.PauliChannel(probX=probX, probY=probY, probZ=probZ), target=qubit)
            for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        probX = _substitute_value(self.probX, **kwargs)
        probY = _substitute_value(self.probY, **kwargs)
        probZ = _substitute_value(self.probZ, **kwargs)

        return PauliChannel(probX=probX, probY=probY, probZ=probZ)

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return PauliChannel(
            probX=_parameter_from_dict(noise["probX"]),
            probY=_parameter_from_dict(noise["probY"]),
            probZ=_parameter_from_dict(noise["probZ"]),
        )


Noise.register_noise(PauliChannel)


class Depolarizing(SingleProbabilisticNoise_34):
    r"""Depolarizing noise channel which transforms a density matrix :math:`\\rho` according to:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho
            + p/3 X \\rho X^{\\dagger}
            + p/3 Y \\rho Y^{\\dagger}
            + p/3 Z \\rho Z^{\\dagger}

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        X = \\left(
                \\begin{matrix}
                    0 & 1 \\\\
                    1 & 0
                \\end{matrix}
            \\right)

        Y = \\left(
                \\begin{matrix}
                    0 & -i \\\\
                    i &  0
                \\end{matrix}
            \\right)

        Z = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & -1
                \\end{matrix}
            \\right)

        p = \\text{probability}

    This noise channel is shown as `DEPO` in circuit diagrams.
    """

    def __init__(self, probability: Union[FreeParameterExpression, float]):
        super().__init__(
            probability=probability,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("DEPO", [probability])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.Depolarizing.construct(target=target[0], probability=self.probability)

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return f"#pragma braket noise depolarizing({self.probability}) {target_qubit}"

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.sqrt(1 - self.probability) * np.eye(2, dtype=complex)
        K1 = np.sqrt(self.probability / 3) * np.array([[0.0, 1.0], [1.0, 0.0]], dtype=complex)
        K2 = np.sqrt(self.probability / 3) * 1j * np.array([[0.0, -1.0], [1.0, 0.0]], dtype=complex)
        K3 = np.sqrt(self.probability / 3) * np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex)
        return [K0, K1, K2, K3]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def depolarizing(target: QubitSetInput, probability: float) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s)
            probability (float): Probability of depolarizing.

        Returns:
            Iterable[Instruction]: `Iterable` of Depolarizing instructions.

        Examples:
            >>> circ = Circuit().depolarizing(0, probability=0.1)
        """
        return [
            Instruction(Noise.Depolarizing(probability=probability), target=qubit)
            for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return Depolarizing(probability=_substitute_value(self._probability, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return Depolarizing(probability=_parameter_from_dict(noise["probability"]))


Noise.register_noise(Depolarizing)


class TwoQubitDepolarizing(SingleProbabilisticNoise_1516):
    r"""Two-Qubit Depolarizing noise channel which transforms a
        density matrix :math:`\\rho` according to:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho + p/15 (
          IX \\rho IX^{\\dagger} + IY \\rho IY^{\\dagger} + IZ \\rho IZ^{\\dagger}
        + XI \\rho XI^{\\dagger} + XX \\rho XX^{\\dagger} + XY \\rho XY^{\\dagger}
        + XZ \\rho XZ^{\\dagger} + YI \\rho YI^{\\dagger} + YX \\rho YX^{\\dagger}
        + YY \\rho YY^{\\dagger} + YZ \\rho YZ^{\\dagger} + ZI \\rho ZI^{\\dagger}
        + ZX \\rho ZX^{\\dagger} + ZY \\rho ZY^{\\dagger} + ZZ \\rho ZZ^{\\dagger})

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        X = \\left(
                \\begin{matrix}
                    0 & 1 \\\\
                    1 & 0
                \\end{matrix}
            \\right)

        Y = \\left(
                \\begin{matrix}
                    0 & -i \\\\
                    i &  0
                \\end{matrix}
            \\right)

        Z = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & -1
                \\end{matrix}
            \\right)

        p = \\text{probability}

    This noise channel is shown as `DEPO` in circuit diagrams.
    """

    def __init__(self, probability: Union[FreeParameterExpression, float]):
        super().__init__(
            probability=probability,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("DEPO", [probability])] * 2,
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.TwoQubitDepolarizing.construct(
            targets=[target[0], target[1]], probability=self.probability
        )

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit_0 = serialization_properties.format_target(int(target[0]))
        target_qubit_1 = serialization_properties.format_target(int(target[1]))
        return (
            f"#pragma braket noise two_qubit_depolarizing({self.probability}) "
            f"{target_qubit_0}, {target_qubit_1}"
        )

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        SI = np.array([[1.0, 0.0], [0.0, 1.0]], dtype=complex)
        SX = np.array([[0.0, 1.0], [1.0, 0.0]], dtype=complex)
        SY = np.array([[0.0, -1.0j], [1.0j, 0.0]], dtype=complex)
        SZ = np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex)

        K_list_single = [SI, SX, SY, SZ]
        K_list = [np.kron(i, j) for i in K_list_single for j in K_list_single]

        K_list[0] *= np.sqrt(1 - self._probability)

        K_list[1:] = [np.sqrt(self._probability / 15) * i for i in K_list[1:]]

        return K_list

    @staticmethod
    def fixed_qubit_count() -> int:
        return 2

    @staticmethod
    @circuit.subroutine(register=True)
    def two_qubit_depolarizing(
        target1: QubitInput, target2: QubitInput, probability: float
    ) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target1 (QubitInput): Target qubit 1.
            target2 (QubitInput): Target qubit 2.
            probability (float): Probability of two-qubit depolarizing.

        Returns:
            Iterable[Instruction]: `Iterable` of Depolarizing instructions.

        Examples:
            >>> circ = Circuit().two_qubit_depolarizing(0, 1, probability=0.1)
        """
        return [
            Instruction(
                Noise.TwoQubitDepolarizing(probability=probability), target=[target1, target2]
            )
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return TwoQubitDepolarizing(probability=_substitute_value(self._probability, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return TwoQubitDepolarizing(probability=_parameter_from_dict(noise["probability"]))


Noise.register_noise(TwoQubitDepolarizing)


class TwoQubitDephasing(SingleProbabilisticNoise_34):
    r"""Two-Qubit Dephasing noise channel which transforms a
        density matrix :math:`\\rho` according to:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho + p/3 (
          IZ \\rho IZ^{\\dagger} + ZI \\rho ZI^{\\dagger} + ZZ \\rho ZZ^{\\dagger})

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        Z = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & -1
                \\end{matrix}
            \\right)

        p = \\text{probability}

    This noise channel is shown as `DEPH` in circuit diagrams.
    """

    def __init__(self, probability: Union[FreeParameterExpression, float]):
        super().__init__(
            probability=probability,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("DEPH", [probability])] * 2,
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.TwoQubitDephasing.construct(
            targets=[target[0], target[1]], probability=self.probability
        )

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit_0 = serialization_properties.format_target(int(target[0]))
        target_qubit_1 = serialization_properties.format_target(int(target[1]))
        return (
            f"#pragma braket noise two_qubit_dephasing({self.probability}) "
            f"{target_qubit_0}, {target_qubit_1}"
        )

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        SI = np.array([[1.0, 0.0], [0.0, 1.0]], dtype=complex)
        SZ = np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex)
        K0 = np.sqrt(1 - self._probability) * np.kron(SI, SI)
        K1 = np.sqrt(self._probability / 3) * np.kron(SI, SZ)
        K2 = np.sqrt(self._probability / 3) * np.kron(SZ, SI)
        K3 = np.sqrt(self._probability / 3) * np.kron(SZ, SZ)

        return [K0, K1, K2, K3]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 2

    @staticmethod
    @circuit.subroutine(register=True)
    def two_qubit_dephasing(
        target1: QubitInput, target2: QubitInput, probability: float
    ) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target1 (QubitInput): Target qubit 1.
            target2 (QubitInput): Target qubit 2.
            probability (float): Probability of two-qubit dephasing.

        Returns:
            Iterable[Instruction]: `Iterable` of Dephasing instructions.

        Examples:
            >>> circ = Circuit().two_qubit_dephasing(0, 1, probability=0.1)
        """
        return [
            Instruction(Noise.TwoQubitDephasing(probability=probability), target=[target1, target2])
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return TwoQubitDephasing(probability=_substitute_value(self._probability, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return TwoQubitDephasing(probability=_parameter_from_dict(noise["probability"]))


Noise.register_noise(TwoQubitDephasing)


class TwoQubitPauliChannel(MultiQubitPauliNoise):
    r"""Two-Qubit Pauli noise channel which transforms a
        density matrix :math:`\\rho` according to:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho +
            p_{IX} IX \\rho IX^{\\dagger} +
            p_{IY} IY \\rho IY^{\\dagger} +
            p_{IZ} IZ \\rho IZ^{\\dagger} +
            p_{XI} XI \\rho XI^{\\dagger} +
            p_{XX} XX \\rho XX^{\\dagger} +
            p_{XY} XY \\rho XY^{\\dagger} +
            p_{XZ} XZ \\rho XZ^{\\dagger} +
            p_{YI} YI \\rho YI^{\\dagger} +
            p_{YX} YX \\rho YX^{\\dagger} +
            p_{YY} YY \\rho YY^{\\dagger} +
            p_{YZ} YZ \\rho YZ^{\\dagger} +
            p_{ZI} ZI \\rho ZI^{\\dagger} +
            p_{ZX} ZX \\rho ZX^{\\dagger} +
            p_{ZY} ZY \\rho ZY^{\\dagger} +
            p_{ZZ} ZZ \\rho ZZ^{\\dagger})

    where

    .. math::
        I = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & 1
                \\end{matrix}
            \\right)

        X = \\left(
                \\begin{matrix}
                    0 & 1 \\\\
                    1 & 0
                \\end{matrix}
            \\right)

        Y = \\left(
                \\begin{matrix}
                    0 & -i \\\\
                    i &  0
                \\end{matrix}
            \\right)

        Z = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & -1
                \\end{matrix}
            \\right)

        p = \\text{sum of all probabilities}

    This noise channel is shown as `PC_2({"pauli_string": probability})` in circuit diagrams.
    """

    _paulis: ClassVar = {
        "I": np.array([[1.0, 0.0], [0.0, 1.0]], dtype=complex),
        "X": np.array([[0.0, 1.0], [1.0, 0.0]], dtype=complex),
        "Y": np.array([[0.0, -1.0j], [1.0j, 0.0]], dtype=complex),
        "Z": np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex),
    }
    _tensor_products_strings = itertools.product(_paulis.keys(), repeat=2)
    _names_list: ClassVar = ["".join(x) for x in _tensor_products_strings]

    def __init__(self, probabilities: dict[str, float]):
        super().__init__(
            probabilities=probabilities,
            qubit_count=None,
            ascii_symbols=[
                f"PC2({probabilities})",
                f"PC2({probabilities})",
            ],
        )
        self._matrix = None

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        if self._matrix is not None:
            return self._matrix
        total_prob = sum(self._probabilities.values())
        K_list = [np.sqrt(1 - total_prob) * np.identity(4)]  # "II" element
        for pstring in self._names_list[1:]:  # ignore "II"
            if pstring in self._probabilities:
                mat = np.sqrt(self._probabilities[pstring]) * np.kron(
                    self._paulis[pstring[0]], self._paulis[pstring[1]]
                )
                K_list.append(mat)
            else:
                K_list.append(np.zeros((4, 4)))
        self._matrix = K_list
        return self._matrix

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.MultiQubitPauliChannel.construct(
            targets=[target[0], target[1]], probabilities=self.probabilities
        )

    @staticmethod
    def fixed_qubit_count() -> int:
        return 2

    @staticmethod
    @circuit.subroutine(register=True)
    def two_qubit_pauli_channel(
        target1: QubitInput, target2: QubitInput, probabilities: dict[str, float]
    ) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target1 (QubitInput): Target qubit 1.
            target2 (QubitInput): Target qubit 2.
            probabilities (dict[str, float]): Probability of two-qubit Pauli channel.

        Returns:
            Iterable[Instruction]: `Iterable` of Depolarizing instructions.

        Examples:
            >>> circ = Circuit().two_qubit_pauli_channel(0, 1, {"XX": 0.1})
        """
        return [
            Instruction(
                Noise.TwoQubitPauliChannel(probabilities=probabilities),
                target=[target1, target2],
            )
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        probabilities = {
            pauli_string: _substitute_value(prob, **kwargs)
            for pauli_string, prob in self._probabilities.items()
        }
        return TwoQubitPauliChannel(probabilities=probabilities)

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        probabilities = {}
        for pauli_string, prob in noise["probabilities"].items():
            probabilities[pauli_string] = _parameter_from_dict(prob)
        return TwoQubitPauliChannel(probabilities=probabilities)


Noise.register_noise(TwoQubitPauliChannel)


class AmplitudeDamping(DampingNoise):
    r"""AmplitudeDamping noise channel which transforms a density matrix :math:`\\rho` according to:

    .. math:: \\rho \\Rightarrow E_0 \\rho E_0^{\\dagger} + E_1 \\rho E_1^{\\dagger}

    where

    .. math::
        E_0 = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & \\sqrt{1-\\gamma}
                \\end{matrix}
              \\right)

        E_1 = \\left(
                \\begin{matrix}
                    0 & \\sqrt{\\gamma} \\\\
                    0 & 0
                \\end{matrix}
              \\right)

    This noise channel is shown as `AD` in circuit diagrams.
    """

    def __init__(self, gamma: Union[FreeParameterExpression, float]):
        super().__init__(
            gamma=gamma,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("AD", [gamma])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.AmplitudeDamping.construct(target=target[0], gamma=self.gamma)

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return f"#pragma braket noise amplitude_damping({self.gamma}) {target_qubit}"

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.array([[1.0, 0.0], [0.0, np.sqrt(1 - self.gamma)]], dtype=complex)
        K1 = np.array([[0.0, np.sqrt(self.gamma)], [0.0, 0.0]], dtype=complex)
        return [K0, K1]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def amplitude_damping(target: QubitSetInput, gamma: float) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s).
            gamma (float): decaying rate of the amplitude damping channel.

        Returns:
            Iterable[Instruction]: `Iterable` of AmplitudeDamping instructions.

        Examples:
            >>> circ = Circuit().amplitude_damping(0, gamma=0.1)
        """
        return [
            Instruction(Noise.AmplitudeDamping(gamma=gamma), target=qubit)
            for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return AmplitudeDamping(gamma=_substitute_value(self._gamma, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return AmplitudeDamping(gamma=_parameter_from_dict(noise["gamma"]))


Noise.register_noise(AmplitudeDamping)


class GeneralizedAmplitudeDamping(GeneralizedAmplitudeDampingNoise):
    r"""Generalized AmplitudeDamping noise channel which transforms a
        density matrix :math:`\\rho` according to:

    .. math:: \\rho \\Rightarrow E_0 \\rho E_0^{\\dagger} + E_1 \\rho E_1^{\\dagger}
                + E_2 \\rho E_2^{\\dagger} + E_3 \\rho E_3^{\\dagger}

    where

    .. math::
        E_0 = \\sqrt(probability)\\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & \\sqrt{1-\\gamma}
                \\end{matrix}
              \\right)

        E_1 = \\sqrt(probability)\\left(
                \\begin{matrix}
                    0 & \\sqrt{\\gamma} \\\\
                    0 & 0
                \\end{matrix}
              \\right)
        E_2 = \\sqrt(1-probability)\\left(
                \\begin{matrix}
                    \\sqrt{1-\\gamma} & 0 \\\\
                    0 & 1
                \\end{matrix}
              \\right)
        E_3 = \\sqrt(1-probability)\\left(
                \\begin{matrix}
                    0 & 0 \\\\
                    \\sqrt{\\gamma} & 0
                \\end{matrix}
              \\right)

    This noise channel is shown as `GAD` in circuit diagrams.
    """

    def __init__(
        self,
        gamma: Union[FreeParameterExpression, float],
        probability: Union[FreeParameterExpression, float],
    ):
        super().__init__(
            gamma=gamma,
            probability=probability,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("GAD", [gamma, probability])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.GeneralizedAmplitudeDamping.construct(
            target=target[0], gamma=self.gamma, probability=self.probability
        )

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return (
            "#pragma braket noise generalized_amplitude_damping("
            f"{self.gamma}, {self.probability}) {target_qubit}"
        )

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.sqrt(self.probability) * np.array(
            [[1.0, 0.0], [0.0, np.sqrt(1 - self.gamma)]], dtype=complex
        )
        K1 = np.sqrt(self.probability) * np.array(
            [[0.0, np.sqrt(self.gamma)], [0.0, 0.0]], dtype=complex
        )
        K2 = np.sqrt(1 - self.probability) * np.array([[np.sqrt(1 - self.gamma), 0.0], [0.0, 1.0]])
        K3 = np.sqrt(1 - self.probability) * np.array([[0.0, 0.0], [np.sqrt(self.gamma), 0.0]])
        return [K0, K1, K2, K3]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def generalized_amplitude_damping(
        target: QubitSetInput, gamma: float, probability: float
    ) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s).
            gamma (float): The damping rate of the amplitude damping channel.
            probability(float): Probability of the system being excited by the environment.

        Returns:
            Iterable[Instruction]: `Iterable` of GeneralizedAmplitudeDamping instructions.

        Examples:
            >>> circ = Circuit().generalized_amplitude_damping(0, gamma=0.1, probability = 0.9)
        """
        return [
            Instruction(
                Noise.GeneralizedAmplitudeDamping(gamma=gamma, probability=probability),
                target=qubit,
            )
            for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        gamma = _substitute_value(self._gamma, **kwargs)
        probability = _substitute_value(self._probability, **kwargs)
        return GeneralizedAmplitudeDamping(gamma=gamma, probability=probability)

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return GeneralizedAmplitudeDamping(
            gamma=_parameter_from_dict(noise["gamma"]),
            probability=_parameter_from_dict(noise["probability"]),
        )


Noise.register_noise(GeneralizedAmplitudeDamping)


class PhaseDamping(DampingNoise):
    r"""Phase damping noise channel which transforms a density matrix :math:`\\rho` according to:

    .. math:: \\rho \\Rightarrow E_0 \\rho E_0^{\\dagger} + E_1 \\rho E_1^{\\dagger}

    where

    .. math::
        E_0 = \\left(
                \\begin{matrix}
                    1 & 0 \\\\
                    0 & \\sqrt{1-gamma}
                \\end{matrix}
              \\right)

        E_1 = \\left(
                \\begin{matrix}
                    0 & 0 \\\\
                    0 & \\sqrt{gamma}
                \\end{matrix}
              \\right)

        p = \\text{probability}

    This noise channel is shown as `PD` in circuit diagrams.
    """

    def __init__(self, gamma: Union[FreeParameterExpression, float]):
        super().__init__(
            gamma=gamma,
            qubit_count=None,
            ascii_symbols=[_ascii_representation("PD", [gamma])],
        )

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.PhaseDamping.construct(target=target[0], gamma=self.gamma)

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        target_qubit = serialization_properties.format_target(int(target[0]))
        return f"#pragma braket noise phase_damping({self.gamma}) {target_qubit}"

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        K0 = np.array([[1.0, 0.0], [0.0, np.sqrt(1 - self.gamma)]], dtype=complex)
        K1 = np.array([[0.0, 0.0], [0.0, np.sqrt(self.gamma)]], dtype=complex)
        return [K0, K1]

    @staticmethod
    def fixed_qubit_count() -> int:
        return 1

    @staticmethod
    @circuit.subroutine(register=True)
    def phase_damping(target: QubitSetInput, gamma: float) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            target (QubitSetInput): Target qubit(s)
            gamma (float): Probability of phase damping.

        Returns:
            Iterable[Instruction]: `Iterable` of PhaseDamping instructions.

        Examples:
            >>> circ = Circuit().phase_damping(0, gamma=0.1)
        """
        return [
            Instruction(Noise.PhaseDamping(gamma=gamma), target=qubit) for qubit in QubitSet(target)
        ]

    def bind_values(self, **kwargs) -> Noise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            Noise: A new Noise object of the same type with the requested
            parameters bound.
        """
        return PhaseDamping(gamma=_substitute_value(self._gamma, **kwargs))

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        return PhaseDamping(gamma=_parameter_from_dict(noise["gamma"]))


Noise.register_noise(PhaseDamping)


class Kraus(Noise):
    """User-defined noise channel that uses the provided matrices as Kraus operators
    This noise channel is shown as `NK` in circuit diagrams.
    """

    def __init__(self, matrices: Iterable[np.ndarray], display_name: str = "KR"):
        """Inits `Kraus`.

        Args:
            matrices (Iterable[ndarray]): A list of matrices that define a noise
                channel. These matrices need to satisfy the requirement of CPTP map.
            display_name (str): Name to be used for an instance of this general noise
                channel for circuit diagrams. Defaults to `KR`.

        Raises:
            ValueError: If any matrix in `matrices` is not a two-dimensional square
                matrix,
                or has a dimension length which is not a positive exponent of 2,
                or the `matrices` do not satisfy CPTP condition.

        """
        for matrix in matrices:
            verify_quantum_operator_matrix_dimensions(matrix)
            if not int(np.log2(matrix.shape[0])) == int(np.log2(matrices[0].shape[0])):
                raise ValueError(f"all matrices in {matrices} must have the same shape")
        self._matrices = [np.array(matrix, dtype=complex) for matrix in matrices]
        self._display_name = display_name
        qubit_count = int(np.log2(self._matrices[0].shape[0]))
        if qubit_count > 2:
            raise ValueError("Kraus operators with more than two qubits are not supported.")
        if len(matrices) > 2 ** (2 * qubit_count):
            raise ValueError("The number of Kraus operators is beyond limit.")

        if not is_cptp(self._matrices):
            raise ValueError(
                "The input matrices do not define a completely-positive trace-preserving map."
            )

        super().__init__(qubit_count=qubit_count, ascii_symbols=[display_name] * qubit_count)

    def to_matrix(self) -> Iterable[np.ndarray]:
        """Returns a matrix representation of this noise.

        Returns:
            Iterable[ndarray]: A list of matrix representations of this noise.
        """
        return self._matrices

    def _to_jaqcd(self, target: QubitSet) -> Any:
        return ir.Kraus.construct(
            targets=list(target),
            matrices=Kraus._transform_matrix_to_ir(self._matrices),
        )

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        matrix_list = ", ".join(
            np.array2string(
                matrix,
                separator=", ",
                formatter={"all": lambda x: format_complex(x)},
            ).replace("\n", "")
            for matrix in self._matrices
        )
        qubit_list = ", ".join(
            serialization_properties.format_target(int(qubit)) for qubit in target
        )
        return f"#pragma braket noise kraus({matrix_list}) {qubit_list}"

    @staticmethod
    def _transform_matrix_to_ir(matrices: Iterable[np.ndarray]) -> list:
        serializable = []
        for matrix in matrices:
            matrix_as_list = [
                [[element.real, element.imag] for element in row] for row in matrix.tolist()
            ]
            serializable.append(matrix_as_list)
        return serializable

    @staticmethod
    @circuit.subroutine(register=True)
    def kraus(
        targets: QubitSetInput, matrices: Iterable[np.array], display_name: str = "KR"
    ) -> Iterable[Instruction]:
        """Registers this function into the circuit class.

        Args:
            targets (QubitSetInput): Target qubit(s)
            matrices (Iterable[array]): Matrices that define a general noise channel.
            display_name (str): The display name.

        Returns:
            Iterable[Instruction]: `Iterable` of Kraus instructions.

        Examples:
            >>> K0 = np.eye(4) * np.sqrt(0.9)
            >>> K1 = np.kron([[1., 0.],[0., 1.]], [[0., 1.],[1., 0.]]) * np.sqrt(0.1)
            >>> circ = Circuit().kraus([1, 0], matrices=[K0, K1])
        """
        if 2 ** len(targets) != matrices[0].shape[0]:
            raise ValueError(
                "Dimensions of the supplied Kraus matrices are incompatible with the targets"
            )

        return Instruction(
            Noise.Kraus(matrices=matrices, display_name=display_name), target=targets
        )

    def to_dict(self) -> dict:
        """Converts this object into a dictionary representation. Not implemented at this time.

        Returns:
            dict: Not implemented at this time..
        """
        raise NotImplementedError

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representation of this class into this class.

        Args:
            noise(dict): The dictionary representation of this noise.

        Returns:
            Noise: A Noise object that represents the passed in dictionary.
        """
        raise NotImplementedError


Noise.register_noise(Kraus)


def _ascii_representation(
    noise: str, parameters: list[Union[FreeParameterExpression, float]]
) -> str:
    """Generates a formatted ascii representation of a noise.

    Args:
        noise (str): The name of the noise.
        parameters (list[Union[FreeParameterExpression, float]]): The parameters to the noise.

    Returns:
        str: The ascii representation of the noise.
    """
    param_list = []
    for param in parameters:
        param_list.append(
            str(param) if isinstance(param, FreeParameterExpression) else f"{param:.2g}"
        )
    param_str = ",".join(param_list)
    return f"{noise}({param_str})"


def _substitute_value(expr: float, **kwargs) -> Union[FreeParameterExpression, float]:
    return expr.subs(kwargs) if isinstance(expr, FreeParameterExpression) else expr


def _parameter_from_dict(parameter: Union[dict, float]) -> Union[FreeParameter, float]:
    """Converts a parameter from a dictionary if it's a FreeParameter, otherwise returns the float.

    Args:
        parameter(Union[dict, float]): The parameter to convert.

    Returns:
        Union[FreeParameter, float]: A FreeParameter representing the parameter, if the parameter
        is a dictionary, otherwise returns the float.
    """
    if isinstance(parameter, dict):
        return FreeParameter.from_dict(parameter)
    return parameter



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import builtins
from copy import deepcopy
from inspect import signature
from typing import Any, Union

from openpulse import ast
from oqpy import BitVar, PhysicalQubits, Program
from sympy import Expr

from braket.parametric.free_parameter import FreeParameter
from braket.parametric.free_parameter_expression import FreeParameterExpression
from braket.parametric.parameterizable import Parameterizable
from braket.pulse.ast.approximation_parser import _ApproximationParser
from braket.pulse.ast.free_parameters import _FreeParameterTransformer
from braket.pulse.ast.qasm_parser import ast_to_qasm
from braket.pulse.ast.qasm_transformer import _IRQASMTransformer
from braket.pulse.frame import Frame
from braket.pulse.pulse_sequence_trace import PulseSequenceTrace
from braket.pulse.waveforms import Waveform
from braket.registers.qubit_set import QubitSet


class PulseSequence:
    """A representation of a collection of instructions to be performed on a quantum device
    and the requested results.
    """

    def __init__(self):
        self._capture_v0_count = 0
        self._program = Program(simplify_constants=False)
        self._frames = {}
        self._waveforms = {}
        self._free_parameters = set()

    def to_time_trace(self) -> PulseSequenceTrace:
        """Generate an approximate trace of the amplitude, frequency, phase for each frame
        contained in the PulseSequence, under the action of the instructions contained in
        the pulse sequence.

        Returns:
            PulseSequenceTrace: The approximation information with each attribute
            (amplitude, frequency and phase) mapping a str (frame id) to a TimeSeries
            (containing the time evolution of that attribute).
        """
        parser = _ApproximationParser(deepcopy(self._program), self._frames)
        return PulseSequenceTrace(
            amplitudes=parser.amplitudes, frequencies=parser.frequencies, phases=parser.phases
        )

    @property
    def parameters(self) -> set[FreeParameter]:
        """Returns the set of `FreeParameter` s in the PulseSequence."""
        return self._free_parameters.copy()

    def set_frequency(
        self, frame: Frame, frequency: Union[float, FreeParameterExpression]
    ) -> PulseSequence:
        """Adds an instruction to set the frequency of the frame to the specified `frequency` value.

        Args:
            frame (Frame): Frame for which the frequency needs to be set.
            frequency (Union[float, FreeParameterExpression]): frequency value to set
                for the specified frame.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        _validate_uniqueness(self._frames, frame)
        self._register_free_parameters(frequency)
        self._program.set_frequency(frame=frame, freq=frequency)
        self._frames[frame.id] = frame
        return self

    def shift_frequency(
        self, frame: Frame, frequency: Union[float, FreeParameterExpression]
    ) -> PulseSequence:
        """Adds an instruction to shift the frequency of the frame by the specified `frequency`
        value.

        Args:
            frame (Frame): Frame for which the frequency needs to be shifted.
            frequency (Union[float, FreeParameterExpression]): frequency value by which to shift
                the frequency for the specified frame.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        _validate_uniqueness(self._frames, frame)
        self._register_free_parameters(frequency)
        self._program.shift_frequency(frame=frame, freq=frequency)
        self._frames[frame.id] = frame
        return self

    def set_phase(
        self, frame: Frame, phase: Union[float, FreeParameterExpression]
    ) -> PulseSequence:
        """Adds an instruction to set the phase of the frame to the specified `phase` value.

        Args:
            frame (Frame): Frame for which the frequency needs to be set.
            phase (Union[float, FreeParameterExpression]): phase value to set
                for the specified frame.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        _validate_uniqueness(self._frames, frame)
        self._register_free_parameters(phase)
        self._program.set_phase(frame=frame, phase=phase)
        self._frames[frame.id] = frame
        return self

    def shift_phase(
        self, frame: Frame, phase: Union[float, FreeParameterExpression]
    ) -> PulseSequence:
        """Adds an instruction to shift the phase of the frame by the specified `phase` value.

        Args:
            frame (Frame): Frame for which the phase needs to be shifted.
            phase (Union[float, FreeParameterExpression]): phase value by which to shift
                the phase for the specified frame.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        _validate_uniqueness(self._frames, frame)
        self._register_free_parameters(phase)
        self._program.shift_phase(frame=frame, phase=phase)
        self._frames[frame.id] = frame
        return self

    def set_scale(
        self, frame: Frame, scale: Union[float, FreeParameterExpression]
    ) -> PulseSequence:
        """Adds an instruction to set the scale on the frame to the specified `scale` value.

        Args:
            frame (Frame): Frame for which the scale needs to be set.
            scale (Union[float, FreeParameterExpression]): scale value to set
                on the specified frame.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        _validate_uniqueness(self._frames, frame)
        self._register_free_parameters(scale)
        self._program.set_scale(frame=frame, scale=scale)
        self._frames[frame.id] = frame
        return self

    def delay(
        self,
        qubits_or_frames: Union[Frame, list[Frame], QubitSet],
        duration: Union[float, FreeParameterExpression],
    ) -> PulseSequence:
        """Adds an instruction to advance the frame clock by the specified `duration` value.

        Args:
            qubits_or_frames (Union[Frame, list[Frame], QubitSet]): Qubits or frame(s) on which
                the delay needs to be introduced.
            duration (Union[float, FreeParameterExpression]): value (in seconds) defining
                the duration of the delay.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        self._register_free_parameters(duration)
        if not isinstance(qubits_or_frames, QubitSet):
            if not isinstance(qubits_or_frames, list):
                qubits_or_frames = [qubits_or_frames]
            _validate_uniqueness(self._frames, qubits_or_frames)
            self._program.delay(time=duration, qubits_or_frames=qubits_or_frames)
            for frame in qubits_or_frames:
                self._frames[frame.id] = frame
        else:
            physical_qubits = [PhysicalQubits[int(x)] for x in qubits_or_frames]
            self._program.delay(time=duration, qubits_or_frames=physical_qubits)
        return self

    def barrier(self, qubits_or_frames: Union[list[Frame], QubitSet]) -> PulseSequence:
        """Adds an instruction to align the frame clocks to the latest time across all the specified
        frames.

        Args:
            qubits_or_frames (Union[list[Frame], QubitSet]): Qubits or frames which the delay
                needs to be introduced.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        if not isinstance(qubits_or_frames, QubitSet):
            _validate_uniqueness(self._frames, qubits_or_frames)
            self._program.barrier(qubits_or_frames=qubits_or_frames)
            for frame in qubits_or_frames:
                self._frames[frame.id] = frame
        else:
            physical_qubits = [PhysicalQubits[int(x)] for x in qubits_or_frames]
            self._program.barrier(qubits_or_frames=physical_qubits)
        return self

    def play(self, frame: Frame, waveform: Waveform) -> PulseSequence:
        """Adds an instruction to play the specified waveform on the supplied frame.

        Args:
            frame (Frame): Frame on which the specified waveform signal would be output.
            waveform (Waveform): Waveform envelope specifying the signal to output on the
                specified frame.

        Returns:
            PulseSequence: returns self.
        """
        _validate_uniqueness(self._frames, frame)
        _validate_uniqueness(self._waveforms, waveform)
        if isinstance(waveform, Parameterizable):
            for param in waveform.parameters:
                self._register_free_parameters(param)
        self._program.play(frame=frame, waveform=waveform)
        self._frames[frame.id] = frame
        self._waveforms[waveform.id] = waveform
        return self

    def capture_v0(self, frame: Frame) -> PulseSequence:
        """Adds an instruction to capture the bit output from measuring the specified frame.

        Args:
            frame (Frame): Frame on which the capture operation needs
                to be performed.

        Returns:
            PulseSequence: self, with the instruction added.
        """
        _validate_uniqueness(self._frames, frame)
        self._program.function_call("capture_v0", [frame])
        self._capture_v0_count += 1
        self._frames[frame.id] = frame
        return self

    def make_bound_pulse_sequence(self, param_values: dict[str, float]) -> PulseSequence:
        """Binds FreeParameters based upon their name and values passed in. If parameters
        share the same name, all the parameters of that name will be set to the mapped value.

        Args:
            param_values (dict[str, float]):  A mapping of FreeParameter names
                to a value to assign to them.

        Returns:
            PulseSequence: Returns a PulseSequence with all present parameters fixed to
            their respective values.
        """
        program = deepcopy(self._program)
        tree: ast.Program = program.to_ast(include_externs=False, ignore_needs_declaration=True)
        new_tree: ast.Program = _FreeParameterTransformer(param_values, program).visit(tree)

        new_program = Program(simplify_constants=False)
        new_program.declared_vars = program.declared_vars
        new_program.undeclared_vars = program.undeclared_vars
        for param_name in param_values:
            new_program.undeclared_vars.pop(param_name, None)
        for x in new_tree.statements:
            new_program._add_statement(x)

        new_pulse_sequence = PulseSequence()
        new_pulse_sequence._program = new_program
        new_pulse_sequence._frames = deepcopy(self._frames)
        new_pulse_sequence._waveforms = {
            wf.id: wf.bind_values(**param_values) if isinstance(wf, Parameterizable) else wf
            for wf in deepcopy(self._waveforms).values()
        }

        # Update waveforms to bind values
        for v in new_program.undeclared_vars:
            if v in self._waveforms:
                new_program.undeclared_vars[v] = new_pulse_sequence._waveforms[
                    v
                ]._to_oqpy_expression()

        new_pulse_sequence._capture_v0_count = self._capture_v0_count
        new_pulse_sequence._free_parameters = {
            p for p in self._free_parameters if p.name not in param_values
        }

        return new_pulse_sequence

    def to_ir(self, sort_input_parameters: bool = False) -> str:
        """Converts this OpenPulse problem into IR representation.

        Args:
            sort_input_parameters (bool): whether input parameters should be printed
                in a sorted order. Defaults to False.

        Returns:
            str: a str representing the OpenPulse program encoding the PulseSequence.
        """
        program = deepcopy(self._program)
        program.autodeclare(encal=False)
        parameters = (
            sorted(self.parameters, key=lambda p: p.name, reverse=True)
            if sort_input_parameters
            else self.parameters
        )
        for param in parameters:
            program.declare(param._to_oqpy_expression(), to_beginning=True)

        if self._capture_v0_count:
            register_identifier = "psb"
            program.declare(
                BitVar[self._capture_v0_count](name=register_identifier), to_beginning=True
            )
            tree = program.to_ast(encal=True, include_externs=False)
            tree = _IRQASMTransformer(register_identifier).visit(tree)
        else:
            tree = program.to_ast(encal=True, include_externs=False)
        return ast_to_qasm(tree)

    def _register_free_parameters(
        self,
        parameter: Union[float, FreeParameterExpression],
    ) -> None:
        if isinstance(parameter, FreeParameterExpression) and isinstance(
            parameter.expression, Expr
        ):
            for p in parameter.expression.free_symbols:
                self._free_parameters.add(FreeParameter(p.name))

    def _parse_arg_from_calibration_schema(
        self, argument: dict, waveforms: dict[Waveform], frames: dict[Frame]
    ) -> Any:
        nonprimitive_arg_type = {
            "frame": frames.get,
            "waveform": waveforms.get,
            "expr": FreeParameterExpression,
        }
        if argument["type"] in nonprimitive_arg_type.keys():
            return nonprimitive_arg_type[argument["type"]](argument["value"])
        else:
            return getattr(builtins, argument["type"])(argument["value"])

    @classmethod
    def _parse_from_calibration_schema(
        cls, calibration: dict, waveforms: dict[Waveform], frames: dict[Frame]
    ) -> PulseSequence:
        """Parsing a JSON input based on https://github.com/aws/amazon-braket-schemas-python/blob/main/src/braket/device_schema/pulse/native_gate_calibrations_v1.py#L26. # noqa: E501

        Args:
            calibration (dict): The pulse instruction to parse
            waveforms (dict[Waveform]): The waveforms supplied for the pulse sequences.
            frames (dict[Frame]): A dictionary of frame objects to use.

        Raises:
            ValueError: If the requested instruction has not been implemented for pulses.

        Returns:
            PulseSequence: The parse sequence obtain from parsing a pulse instruction.
        """
        calibration_sequence = cls()
        for instr in calibration:
            if hasattr(PulseSequence, f"{instr['name']}"):
                instr_function = getattr(calibration_sequence, instr["name"])
                instr_args_keys = signature(instr_function).parameters.keys()
                instr_args = {}
                if instr["arguments"] is not None:
                    for argument in instr["arguments"]:
                        if argument["name"] in {"qubit", "frame"} and instr["name"] in {
                            "barrier",
                            "delay",
                        }:
                            argument_value = (
                                [frames[argument["value"]]]
                                if argument["name"] == "frame"
                                else instr_args.get("qubits_or_frames", QubitSet())
                            )
                            # QubitSet is an IndexedSet so the ordering matters
                            if argument["name"] == "frame":
                                argument_value = (
                                    instr_args.get("qubits_or_frames", []) + argument_value
                                )
                            else:
                                argument_value.update(QubitSet(int(argument["value"])))
                            instr_args["qubits_or_frames"] = argument_value
                        elif argument["name"] in instr_args_keys:
                            instr_args[argument["name"]] = (
                                calibration_sequence._parse_arg_from_calibration_schema(
                                    argument, waveforms, frames
                                )
                            )
                else:
                    instr_args["qubits_or_frames"] = []
                instr_function(**instr_args)
            else:
                raise ValueError(f"The {instr['name']} instruction has not been implemented")
        return calibration_sequence

    def __call__(
        self, arg: Any | None = None, **kwargs: Union[FreeParameter, str]
    ) -> PulseSequence:
        """Implements the call function to easily make a bound PulseSequence.

        Args:
            arg (Any | None): A value to bind to all parameters. Defaults to None and
                can be overridden if the parameter is in kwargs.
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            PulseSequence: A pulse sequence with the specified parameters bound.
        """
        param_values = {}
        if arg is not None:
            for param in self.parameters:
                param_values[str(param)] = arg
        for key, val in kwargs.items():
            param_values[str(key)] = val
        return self.make_bound_pulse_sequence(param_values)

    def __eq__(self, other: PulseSequence):
        sort_input_parameters = True
        return (
            isinstance(other, PulseSequence)
            and self.parameters == other.parameters
            and self.to_ir(sort_input_parameters) == other.to_ir(sort_input_parameters)
        )


def _validate_uniqueness(
    mapping: dict[str, Any], values: Union[Frame, Waveform, list[Frame], list[Waveform]]
) -> None:
    if not isinstance(values, list):
        values = [values]

    for value in values:
        if value.id in mapping and mapping[value.id] != value:
            raise ValueError(
                f"{value.id} has already been used for defining {mapping[value.id]} "
                f"which differs from {value}"
            )



import json
import os
import tempfile
from pathlib import Path
from unittest.mock import patch

from braket.jobs import (
    get_checkpoint_dir,
    get_hyperparameters,
    get_input_data_dir,
    get_job_device_arn,
    get_job_name,
    get_results_dir,
)


def test_job_name():
    assert get_job_name() == ""
    job_name = "my_job_name"
    with patch.dict(os.environ, {"AMZN_BRAKET_JOB_NAME": job_name}):
        assert get_job_name() == job_name


def test_job_device_arn():
    assert get_job_device_arn() == "local:none/none"
    device_arn = "my_device_arn"
    with patch.dict(os.environ, {"AMZN_BRAKET_DEVICE_ARN": device_arn}):
        assert get_job_device_arn() == device_arn


def test_input_data_dir():
    assert get_input_data_dir() == "."
    input_path = "my/input/path"
    with patch.dict(os.environ, {"AMZN_BRAKET_INPUT_DIR": input_path}):
        assert get_input_data_dir() == f"{input_path}/input"
        channel_name = "my_channel"
        assert get_input_data_dir(channel_name) == f"{input_path}/{channel_name}"


def test_results_dir():
    assert get_results_dir() == "."
    results_dir = "my_results_dir"
    with patch.dict(os.environ, {"AMZN_BRAKET_JOB_RESULTS_DIR": results_dir}):
        assert get_results_dir() == results_dir


def test_checkpoint_dir():
    assert get_checkpoint_dir() == "."
    checkpoint_dir = "my_checkpoint_dir"
    with patch.dict(os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": checkpoint_dir}):
        assert get_checkpoint_dir() == checkpoint_dir


def test_hyperparameters():
    assert get_hyperparameters() == {}
    hp_file = "my_hyperparameters.json"
    hyperparameters = {
        "a": "a_val",
        "b": 2,
    }
    with tempfile.TemporaryDirectory() as temp_dir, patch.dict(
        os.environ, {"AMZN_BRAKET_HP_FILE": str(Path(temp_dir) / hp_file)}
    ):
        with open(str(Path(temp_dir) / hp_file), "w") as f:
            json.dump(hyperparameters, f)
        assert get_hyperparameters() == hyperparameters



"""Sphinx configuration."""

import datetime

import pkg_resources

# Sphinx configuration below.
project = "amazon-braket-sdk"
version = pkg_resources.require(project)[0].version
release = version
copyright = "{}, Amazon.com".format(datetime.datetime.now().year)

extensions = [
    "sphinxcontrib.apidoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
]

source_suffix = ".rst"
master_doc = "index"

autoclass_content = "both"
autodoc_member_order = "bysource"
default_role = "py:obj"

html_theme = "sphinx_rtd_theme"
htmlhelp_basename = "{}doc".format(project)

language = "en"

napoleon_use_rtype = False

apidoc_module_dir = "../src/braket"
apidoc_output_dir = "_apidoc"
apidoc_excluded_paths = ["../test"]
apidoc_separate_modules = True
apidoc_module_first = True
apidoc_extra_args = ["-f", "--implicit-namespaces", "-H", "API Reference"]



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Union

from braket.parametric.free_parameter import FreeParameter
from braket.parametric.free_parameter_expression import FreeParameterExpression


class Parameterizable(ABC):
    """A parameterized object is the abstract definition of an object
    that can take in FreeParameterExpressions.
    """

    @property
    @abstractmethod
    def parameters(self) -> list[Union[FreeParameterExpression, FreeParameter, float]]:
        """Get the parameters.

        Returns:
            list[Union[FreeParameterExpression, FreeParameter, float]]: The parameters associated
            with the object, either unbound free parameter expressions or bound values. The order
            of the parameters is determined by the subclass.
        """

    @abstractmethod
    def bind_values(self, **kwargs: Union[FreeParameter, str]) -> Any:
        """Takes in parameters and returns an object with specified parameters
        replaced with their values.

        Args:
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            Any: The result object will depend on the implementation of the object being bound.
        """



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from enum import Enum


class Devices:
    class _Amazon(str, Enum):
        SV1 = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
        TN1 = "arn:aws:braket:::device/quantum-simulator/amazon/tn1"
        DM1 = "arn:aws:braket:::device/quantum-simulator/amazon/dm1"

    class _DWave(str, Enum):
        _Advantage1 = "arn:aws:braket:::device/qpu/d-wave/Advantage_system1"
        _Advantage3 = "arn:aws:braket:::device/qpu/d-wave/Advantage_system3"
        _Advantage4 = "arn:aws:braket:::device/qpu/d-wave/Advantage_system4"
        _Advantage6 = "arn:aws:braket:us-west-2::device/qpu/d-wave/Advantage_system6"
        _DW2000Q6 = "arn:aws:braket:::device/qpu/d-wave/DW_2000Q_6"

    class _IonQ(str, Enum):
        Harmony = "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony"
        Aria1 = "arn:aws:braket:us-east-1::device/qpu/ionq/Aria-1"
        Aria2 = "arn:aws:braket:us-east-1::device/qpu/ionq/Aria-2"
        Forte1 = "arn:aws:braket:us-east-1::device/qpu/ionq/Forte-1"

    class _OQC(str, Enum):
        Lucy = "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy"

    class _QuEra(str, Enum):
        Aquila = "arn:aws:braket:us-east-1::device/qpu/quera/Aquila"

    class _Rigetti(str, Enum):
        _Aspen8 = "arn:aws:braket:::device/qpu/rigetti/Aspen-8"
        _Aspen9 = "arn:aws:braket:::device/qpu/rigetti/Aspen-9"
        _Aspen10 = "arn:aws:braket:::device/qpu/rigetti/Aspen-10"
        _Aspen11 = "arn:aws:braket:::device/qpu/rigetti/Aspen-11"
        _AspenM1 = "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-1"
        _AspenM2 = "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-2"
        AspenM3 = "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3"

    class _Xanadu(str, Enum):
        _Borealis = "arn:aws:braket:us-east-1::device/qpu/xanadu/Borealis"

    Amazon = _Amazon
    # DWave = _DWave
    IonQ = _IonQ
    OQC = _OQC
    QuEra = _QuEra
    Rigetti = _Rigetti
    # Xanadu = _Xanadu



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from decimal import Decimal
from typing import Optional

from braket.ahs.discretization_types import DiscretizationError
from braket.ahs.pattern import Pattern
from braket.timings.time_series import TimeSeries


class Field:
    def __init__(self, time_series: TimeSeries, pattern: Optional[Pattern] = None) -> None:
        """A space and time dependent parameter of a program.

        Args:
            time_series (TimeSeries): The time series representing this field.
            pattern (Optional[Pattern]): The local pattern of real numbers.
        """
        self._time_series = time_series
        self._pattern = pattern

    @property
    def time_series(self) -> TimeSeries:
        """TimeSeries: The time series representing this field."""
        return self._time_series

    @property
    def pattern(self) -> Optional[Pattern]:
        """Optional[Pattern]: The local pattern of real numbers."""
        return self._pattern

    def discretize(
        self,
        time_resolution: Decimal,
        value_resolution: Decimal,
        pattern_resolution: Optional[Decimal] = None,
    ) -> Field:
        """Creates a discretized version of the field,
        where time, value and pattern are rounded to the
        closest multiple of their corresponding resolutions.

        Args:
            time_resolution (Decimal): Time resolution
            value_resolution (Decimal): Value resolution
            pattern_resolution (Optional[Decimal]): Pattern resolution

        Returns:
            Field: A new discretized field.

        Raises:
            ValueError: if pattern_resolution is None, but there is a Pattern
        """
        discretized_time_series = self.time_series.discretize(time_resolution, value_resolution)
        if self.pattern is None:
            discretized_pattern = None
        else:
            if pattern_resolution is None:
                raise DiscretizationError(
                    f"{self.pattern} is defined but has no pattern_resolution defined"
                )
            discretized_pattern = self.pattern.discretize(pattern_resolution)
        discretized_field = Field(time_series=discretized_time_series, pattern=discretized_pattern)
        return discretized_field



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import math
import random

import pytest
from gate_model_device_testing_utils import bell_pair_openqasm_testing, no_result_types_testing

from braket.aws import AwsDevice
from braket.circuits import Circuit

SHOTS = 100
TN1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/tn1"
SIMULATOR_ARNS = [TN1_ARN]


@pytest.mark.parametrize("simulator_arn", SIMULATOR_ARNS)
def test_ghz(simulator_arn, aws_session, s3_destination_folder):
    num_qubits = 50
    circuit = _ghz(num_qubits)
    device = AwsDevice(simulator_arn, aws_session)
    no_result_types_testing(
        circuit,
        device,
        {"shots": SHOTS, "s3_destination_folder": s3_destination_folder},
        {"0" * num_qubits: 0.5, "1" * num_qubits: 0.5},
    )


@pytest.mark.parametrize("simulator_arn", SIMULATOR_ARNS)
def test_qft_iqft_h(simulator_arn, aws_session, s3_destination_folder):
    num_qubits = 24
    h_qubit = random.randint(0, num_qubits - 1)
    circuit = _inverse_qft(_qft(Circuit().h(h_qubit), num_qubits), num_qubits)
    device = AwsDevice(simulator_arn, aws_session)
    no_result_types_testing(
        circuit,
        device,
        {"shots": SHOTS, "s3_destination_folder": s3_destination_folder},
        {"0" * num_qubits: 0.5, "0" * h_qubit + "1" + "0" * (num_qubits - h_qubit - 1): 0.5},
    )


@pytest.mark.parametrize("simulator_arn", SIMULATOR_ARNS)
def test_bell_pair_openqasm(simulator_arn, aws_session, s3_destination_folder):
    device = AwsDevice(simulator_arn, aws_session)
    bell_pair_openqasm_testing(
        device, {"shots": SHOTS, "s3_destination_folder": s3_destination_folder}
    )


def _ghz(num_qubits):
    circuit = Circuit()
    circuit.h(0)
    for qubit in range(num_qubits - 1):
        circuit.cnot(qubit, qubit + 1)
    return circuit


def _qft(circuit, num_qubits):
    for i in range(num_qubits):
        circuit.h(i)
        for j in range(1, num_qubits - i):
            circuit.cphaseshift(i + j, i, math.pi / (2**j))

    for qubit in range(math.floor(num_qubits / 2)):
        circuit.swap(qubit, num_qubits - qubit - 1)

    return circuit


def _inverse_qft(circuit, num_qubits):
    for qubit in range(math.floor(num_qubits / 2)):
        circuit.swap(qubit, num_qubits - qubit - 1)

    for i in reversed(range(num_qubits)):
        for j in reversed(range(1, num_qubits - i)):
            circuit.cphaseshift(i + j, i, -math.pi / (2**j))
        circuit.h(i)

    return circuit



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from decimal import Decimal
from numbers import Number


class Pattern:
    def __init__(self, series: list[Number]):
        """Represents the spatial dependence of a Field.

        Args:
            series (list[Number]): A series of numbers representing the the local
                pattern of real numbers.
        """
        self._series = series

    @property
    def series(self) -> list[Number]:
        """list[Number]: A series of numbers representing the local
        pattern of real numbers.
        """
        return self._series

    def discretize(self, resolution: Decimal) -> Pattern:
        """Creates a discretized version of the pattern,
        where each value is rounded to the closest multiple
        of the resolution.

        Args:
            resolution (Decimal): Resolution of the discretization

        Returns:
            Pattern: The new discretized pattern
        """
        discretized_series = [round(Decimal(num) / resolution) * resolution for num in self.series]
        return Pattern(series=discretized_series)
